/*
 * Copyright (c) 2008, Koji Tsukada<tsuka@mobiquitous.com> 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution.
 *    
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Collections.Generic;
using System.Text;
using System.Collections;

namespace MobiServer.Devices.X10.Data
{
    public class HouseCodes{
        
        static Hashtable houseCodes = null;

        public static void InitHashtable(){
            houseCodes = new Hashtable();
            houseCodes.Add('A', (byte)0x46);
            houseCodes.Add('B', (byte)0x4E);
            houseCodes.Add('C', (byte)0x42);
            houseCodes.Add('D', (byte)0x4A);
            houseCodes.Add('E', (byte)0x41);
            houseCodes.Add('F', (byte)0x49);
            houseCodes.Add('G', (byte)0x45);
            houseCodes.Add('H', (byte)0x4D);
            houseCodes.Add('I', (byte)0x47);
            houseCodes.Add('J', (byte)0x4F);
            houseCodes.Add('K', (byte)0x43);
            houseCodes.Add('L', (byte)0x4B);
            houseCodes.Add('M', (byte)0x40);
            houseCodes.Add('N', (byte)0x48);
            houseCodes.Add('O', (byte)0x44);
            houseCodes.Add('P', (byte)0x4C);
        }

        
        public static byte GetBinary(char hcode)
        {
            byte result = 0x00;
            hcode = char.ToUpper(hcode);
            foreach (char c in houseCodes.Keys)
            {
                if (c == hcode)
                {
                    result = (byte)houseCodes[c];
                    break;
                }
            }
            return result;
        }

        public static char GetChar(byte hbinary)
        {
            char result = (char)0x00;

            try
            {
                foreach (char c in houseCodes.Keys)
                {
                    byte b = (byte)houseCodes[c];
                    if (b == hbinary)
                    {
                        result = (char)c;
                        break;
                    }
                }
            }
            catch (Exception ex)
            {
                return result;
            }

            return result;
        }
    }

    public class UnitCodes
    {
        static Hashtable unitCodes = null;

        static public void InitHashtable()
        {
            unitCodes = new Hashtable();
            unitCodes.Add(1, (byte)0x4C);
            unitCodes.Add(2, (byte)0x5C);
            unitCodes.Add(3, (byte)0x44);
            unitCodes.Add(4, (byte)0x54);
            unitCodes.Add(5, (byte)0x42);
            unitCodes.Add(6, (byte)0x52);
            unitCodes.Add(7, (byte)0x4A);
            unitCodes.Add(8, (byte)0x5A);
            unitCodes.Add(9, (byte)0x4E);
            unitCodes.Add(10, (byte)0x5E);
            unitCodes.Add(11, (byte)0x46);
            unitCodes.Add(12, (byte)0x56);
            unitCodes.Add(13, (byte)0x40);
            unitCodes.Add(14, (byte)0x50);
            unitCodes.Add(15, (byte)0x48);
            unitCodes.Add(16, (byte)0x58);
        }


        public static byte GetBinary(int ucode)
        {
            byte result = 0x00;
            //code = char.ToUpper(code);
            foreach (int c in unitCodes.Keys)
            {
                if (c == ucode)
                {
                    result = (byte)unitCodes[c];
                    break;
                }
            }
            return result;
        }

        public static int GetNumber(byte ubinary)
        {
            int result = 0;
            foreach (int i in unitCodes.Keys)
            {
                byte b = (byte)unitCodes[i];
                if (b == ubinary)
                {
                    result = (int)i;
                    break;
                }
            }

            return result;
        }

    }

    public class RepeatCounts
    {
        static Hashtable repeatCounts = null;

       public  static void InitHashtable()
        {
            repeatCounts = new Hashtable();
            repeatCounts.Add(1, (byte)0x41);
            repeatCounts.Add(2, (byte)0x42);
            repeatCounts.Add(3, (byte)0x43);
            repeatCounts.Add(4, (byte)0x44);
            repeatCounts.Add(5, (byte)0x45);
            repeatCounts.Add(6, (byte)0x46);
            repeatCounts.Add(7, (byte)0x47);
            repeatCounts.Add(8, (byte)0x48);
            repeatCounts.Add(9, (byte)0x49);
            repeatCounts.Add(10, (byte)0x4A);
            repeatCounts.Add(11, (byte)0x4B);
            repeatCounts.Add(12, (byte)0x4C);
            repeatCounts.Add(13, (byte)0x4D);
            repeatCounts.Add(14, (byte)0x4E);
            repeatCounts.Add(15, (byte)0x4F);
        }


        public static byte GetBinary(int rcount)
        {
            byte result = 0x00;
            //code = char.ToUpper(code);
            foreach (int c in repeatCounts.Keys)
            {
                if (c == rcount)
                {
                    result = (byte)repeatCounts[c];
                    break;
                }
            }
            return result;
        }

        public static int GetNumber(byte rbinary)
        {
            int result = 0;
            foreach (int i in repeatCounts.Keys)
            {
                byte b = (byte)repeatCounts[i];
                if (b == rbinary)
                {
                    result = (int)i;
                    break;
                }
            }
            return result;
        }
    }
    

    public class Commands{
        //8 data bits, no parity, 1 stop bit, 9600 baud
        public const byte START_BIT = 0x02;
        public const byte ACK = 0x06; //followed with CR
        public const byte NAK = 0x15; //NOT followed with CR
        public const byte CR = 0x0D;

        public const byte START_COMMAND = 0x63;
        public const byte ON = 0x45;
        public const byte OFF = 0x47;
        public const byte DIM = 0x49;
        public const byte BRIGHT = 0x4B;

        public const byte ALL_LIGHTS_OFF = 0x4D;
        public const byte ALL_LIGHTS_ON = 0x43;
        public const byte ALL_UNITS_OFF = 0x41;

        public const byte HAIL_REQUEST = 0x51;
        public const byte PRESET_DIM_HIGH = 0x57;
        public const byte PRESET_DIM_LOW = 0x55;
        public const byte STATUS_REQUEST = 0x5F;
        public const byte STATUS_ON = 0x5B;
        public const byte STATUS_OFF = 0x5D;

        public const byte EXTENDED_CODE = 0x4F;
        public const byte EXTENDED_DATA = 0x59;

        public const byte COMMAND_RECEIVED = 0x58;

        public const byte COMMAND_EXTERNAL = 0x78;
    
        public enum ReadyStates { ACK, NAK, Error };


    }

    public class States
    {
        public const string In = "In";
        public const string Out = "Out";
        public const string Attach = "Attach";
        public const string Dettach = "Dettach";
    }

    public class Types{
        public const string Lamp = "Lamp";
        public const string Appliance = "Appliance";
        public const string Power = "Power";
        public const string Bright = "Bright";
    }

    public class Base
    {
        public const string Header = "X10";
        string state;
        char houseCode;
        int unitCode;
        string type;
        //bool flag;
        int data;

        public Base(string state, char houseCode, int unitCode, string type, int data)//bool flag
        {
            this.state = state;
            this.houseCode = houseCode;
            this.unitCode = unitCode;
            this.type = type;
            this.data = data;
        }

        public Base(string state, char houseCode, int unitCode, string type)
        {
            this.state = state;
            this.houseCode = houseCode;
            this.unitCode = unitCode;
            this.type = type;
        }

        public Base(string state, char houseCode, int unitCode)
        {
            this.state = state;
            this.houseCode = houseCode;
            this.unitCode = unitCode;
            this.type = Types.Lamp;
        }

        public string State
        {
            get { return this.state; }
        }

        public char HouseCode
        {
            get { return this.houseCode; }
        }

        public int UnitCode
        {
            get { return this.unitCode; }
        }

        public string Type
        {
            get { return this.type; }
        }

        public int Data
        {
            get { return this.data; }
        }

        public string[] ToArray()
        {
            string[] args = null;

            switch(State){
                case States.Attach:
                case States.Dettach:
                    args = new string[5] { Header, State, HouseCode.ToString(), UnitCode.ToString(), Type };
                    break;
                case States.In:
                case States.Out:
                    args = new string[6] { Header, State, HouseCode.ToString(), UnitCode.ToString(), Type, Data.ToString() };
                    break;
            }
            return args;
        }
        static public Base Parse(string [] args)
        {
            try
            {
                Base b = null;

                if(args[0] != Header)
                    return null;

                string state = args[1];
                char houseCode = args[2].ToCharArray()[0];
                int unitCode = int.Parse(args[3]);
               
                if(args.Length < 5){
                     b = new Base(state, houseCode, unitCode);
                }
                else if (args.Length < 6)
                {
                    string type = args[4];
                    b = new Base(state, houseCode, unitCode, type);
                }
                else
                {
                    string type = args[4];
                    int data = int.Parse(args[5]);
                    b = new Base(state, houseCode, unitCode, type, data);
                }
                return b;
            }
            catch (Exception ex)
            {
                return null;
            }

        }
    }
}

