/*
 * Copyright (c) 2008, Koji Tsukada<tsuka@mobiquitous.com> 
 * All rights reserved.
 * 
 * Redistribution and use in source and binary forms, with or without modification, 
 * are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, 
 *    this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright notice, 
 *    this list of conditions and the following disclaimer in the documentation 
 *    and/or other materials provided with the distribution.
 *    
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND 
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. 
 * IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, 
 * INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, 
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, 
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF 
 * LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE 
 * OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED 
 * OF THE POSSIBILITY OF SUCH DAMAGE.
*/
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
//using System.Windows.Forms;

using System.Collections;
using System.Text.RegularExpressions;
using System.IO;
using System.Threading;

namespace MobiServer.Devices.X10.Control
{
    public partial class Panel : MobiServer.Devices.ControlBase
    {

        #region Constants/Variables
        Hashtable controlHolder = null;
        Control.Base targetControl = null;
       
        int targetHeight = 0;
        const string SETTING_FILE = @"setting\setting_x10.txt";

        public MobiServer.MessageHandler messageHandler;
        public MobiServer.WriteHandler serverWriteHandler;
        string receivedStr = "";
        
        Hashtable serialReadHandlerHolder;
        bool initialized = false;
        Thread startThread = null;
        const int START_WAIT = 1000;

        const string STR_OPEN_PORT = "|[g{0}J܂D";
        const string STR_CLOSE_PORT = "|[g{0}܂D";
        const string STR_ERROR_DEVICETYPE = "foCX̎ނsłD";
        #endregion

        #region Constructor/Destructor
        public Panel()
        {
            InitializeComponent();
        }

        public Panel(MobiServer.TCPServer server, MobiServer.Message message)
            : base(Data.Base.Header, Data.Base.Header, server, message)
        {
            InitializeComponent();
            Data.HouseCodes.InitHashtable();
            Data.UnitCodes.InitHashtable();
            Data.RepeatCounts.InitHashtable();

            controlHolder = new Hashtable();
            LoadSetting();            
        }

        private void Panel_Load(object sender, EventArgs e)
        {
            if (checkBoxAutoStart.Checked)
            {
                OpenPort();
            }
            serverWriteHandler += new MobiServer.WriteHandler(serverControl.OnServerWriteEvent);
            serverControl.AddReadHandler(new MobiServer.ReadHandler(OnServerReadEvent));
            serverControl.messageHandler += new MobiServer.MessageHandler(messageControl.OnMessageEvent);
        
            try
            {
                Hashtable h = X10.Control.Base.LoadX10All();
                foreach (Base b in h.Values)
                {
                    StartControl(b);
                }
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
            initialized = true;
        }

      
        void ChangeGUIEnabled(bool flag)
        {
            groupBoxButton.Enabled = flag;
        }
        #endregion

        #region Serial(/X10) In/Out

        private void buttonOpen_Click(object sender, EventArgs e)
        {
            OpenPort();
        }

        private void buttonClose_Click(object sender, EventArgs e)
        {
            ClosePort();
        }

        void OpenPort()
        {
            buttonOpen.Enabled = false;
            try
            {
                if (!serialPort1.IsOpen)
                {
                    serialPort1.PortName = this.comboBoxComPort.SelectedItem.ToString();
                    serialPort1.DataReceived += new System.IO.Ports.SerialDataReceivedEventHandler(OnDataReceived);
                    serialPort1.Open();
                }
                buttonClose.Enabled = true;
                PrintMessage(STR_OPEN_PORT, serialPort1.PortName);
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
                buttonOpen.Enabled = true;
                buttonClose.Enabled = false;
            }

        }

        void ClosePort()
        {
            buttonClose.Enabled = false;
            try
            {
                if (serialPort1.IsOpen)
                {
                    serialPort1.Close();
                    serialPort1.DataReceived += null;
                }
                buttonOpen.Enabled = true;
                PrintMessage(STR_CLOSE_PORT, serialPort1.PortName);
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
                buttonOpen.Enabled = false;
                buttonClose.Enabled = true;
            }
        }

        public void CreateSerialReadEvent(char houseCode, int unitCode, byte functionCode)
        {
            Data.X10EventArgs e = new MobiServer.Devices.X10.Data.X10EventArgs(houseCode, unitCode, functionCode);
            ChangeGUIEnabled(true);
            foreach (Hashtable lineHolder in controlHolder.Values)
            {
                foreach (Base control in lineHolder.Values)
                {
                    Data.ReadHandler handler = control.serialReadHandler;
                    if (handler != null)
                    {
                        handler.Invoke(this, e);
                    }
                }

            }
            return;
        }

        void OnSerialWriteEvent(object sender, Data.X10EventArgs e)
        {
            SendCommand(e.HouseCode, e.UnitCode, e.FunctionCode, e.RepeatCount);
        }

        void ReadSerialBytes(ref byte[] bytes, int start)
        {
            int i = start;
            while (true)
            {
                int length = serialPort1.BytesToRead;
                serialPort1.Read(bytes, i, length);
                i = i + length;

                if (i >= bytes.Length)
                    break;
            }
        }

        void OnDataReceived(object sender, System.IO.Ports.SerialDataReceivedEventArgs e)
        {
            try
            {
                if (e.EventType == System.IO.Ports.SerialData.Chars)
                {
                    if (serialPort1.BytesToRead == 0)
                        return;

                    byte b = (byte)serialPort1.ReadByte();
                    if (b == Data.Commands.START_COMMAND)
                    {
                        byte[] bytes = new byte[4];
                        ReadSerialBytes(ref bytes, 1);

                        //CreateSerialReadEvent(hcode, ucode, fcode);
                        string str = BitConverter.ToString(bytes);
                        PrintDebug(str);

                    }
                    else if (b == Data.Commands.COMMAND_EXTERNAL ||
                        b == Data.Commands.COMMAND_RECEIVED)
                    {
                        byte[] bytes = new byte[5];
                        byte[] bytes2 = new byte[5];

                        bytes[0] = b;

                        ReadSerialBytes(ref bytes, 1);

                        if (Data.UnitCodes.GetNumber(bytes[2]) == 0)
                        {
                            char hcode = Data.HouseCodes.GetChar(bytes[1]);
                            int ucode = 0;
                            byte fcode = bytes[2];
                            CreateSerialReadEvent(hcode, ucode, fcode);
                            string str = BitConverter.ToString(bytes);
                            PrintDebug(str);

                            if (b == Data.Commands.COMMAND_EXTERNAL)
                            {

                                CreateServerWriteEvent(hcode, ucode, fcode);

                            }
                        }
                        else
                        {
                            ReadSerialBytes(ref bytes2, 0);
                            char hcode = Data.HouseCodes.GetChar(bytes[1]);
                            int ucode = Data.UnitCodes.GetNumber(bytes[2]);
                            byte fcode = bytes2[2];
                            CreateSerialReadEvent(hcode, ucode, fcode);
                            string str = BitConverter.ToString(bytes) + ":" + BitConverter.ToString(bytes2);
                            PrintDebug(str);

                            if (b == Data.Commands.COMMAND_EXTERNAL)
                            {
                                CreateServerWriteEvent(hcode, ucode, fcode);
                            }
                        }
                    }

                    else
                    {
                        int length = serialPort1.BytesToRead;
                        byte[] bytes = new byte[length];

                        serialPort1.Read(bytes, 0, bytes.Length);
                        string str = BitConverter.ToString(bytes);
                        PrintDebug(str);
                    }
                }
                else if (e.EventType == System.IO.Ports.SerialData.Eof)
                {
                }
                else
                {
                    int length = serialPort1.BytesToRead;
                    byte[] bytes = new byte[length];

                    serialPort1.Read(bytes, 0, bytes.Length);
                    string str = BitConverter.ToString(bytes);
                    PrintDebug(str);
                }
            }
            catch (Exception ex)
            {
                PrintDebug(ex.Message);
                ChangeGUIEnabled(true);
            }
        }
        #endregion

        #region X10 In/Out
        Data.Commands.ReadyStates IsX10Ready()
        {
            Data.Commands.ReadyStates state = Data.Commands.ReadyStates.Error;
            try
            {
                if (serialPort1 != null && !serialPort1.IsOpen)
                {
                    serialPort1.PortName = comboBoxComPort.SelectedItem.ToString();
                    serialPort1.Open();
                }

                if (serialPort1 != null && serialPort1.IsOpen)// && serialWriteReady
                {
                    byte b;
                    serialPort1.DataReceived += null;

                    serialPort1.Write(new byte[] { Data.Commands.START_BIT }, 0, 1);
                    b = (byte)serialPort1.ReadByte();
                    if (Data.Commands.ACK == b)
                    {
                        state = Data.Commands.ReadyStates.ACK;
                        b = (byte)serialPort1.ReadByte(); //CR
                    }
                    else if (Data.Commands.NAK == b)
                    {
                        state = Data.Commands.ReadyStates.NAK;
                    }
                    else
                    {
                        state = Data.Commands.ReadyStates.Error;
                    }
                }
            }
            catch (Exception ex)
            {
                // PrintDebug(ex.Message);
                state = Data.Commands.ReadyStates.Error;
            }
            finally
            {
                if (serialPort1 != null && serialPort1.IsOpen)
                {
                    serialPort1.DataReceived += new System.IO.Ports.SerialDataReceivedEventHandler(OnDataReceived);
                }
            }
            return state;
        }

        bool SendCommand(char houseCode, int unitCode, byte functionCode, int repeatCount)
        {
            byte[] bytes = new byte[5];
            bool result = false;

            ChangeGUIEnabled(false);
            lock (this)
            {
                try
                {                   
                    while (true)
                    {
                        if (Data.Commands.ReadyStates.ACK == IsX10Ready())
                            break;

                        Thread.Sleep(100);
                        System.Windows.Forms.Application.DoEvents();
                    }

                    bytes[0] = Data.Commands.START_COMMAND;
                    bytes[1] = Data.HouseCodes.GetBinary(houseCode);

                    if (unitCode != 0)
                    {
                        bytes[2] = Data.UnitCodes.GetBinary(unitCode);
                        bytes[3] = functionCode;
                    }
                    else
                    {
                        bytes[2] = functionCode;
                        bytes[3] = (byte)0x00;
                    }

                    bytes[4] = Data.RepeatCounts.GetBinary(repeatCount);

                    if (serialPort1 != null && serialPort1.IsOpen)
                    {
                        serialPort1.Write(bytes, 0, bytes.Length);
                    }
                    result = true;
                }
                catch (Exception ex)
                {

                }
            }
            return result;
        }

        private void buttonAllLightOn_Click(object sender, EventArgs e)
        {
            SendAllLightOn(comboBoxHouseCode.SelectedItem.ToString().ToCharArray()[0]);
        }

        private void buttonAllLightBright_Click(object sender, EventArgs e)
        {
            SendAllLightBright(comboBoxHouseCode.SelectedItem.ToString().ToCharArray()[0]);
        }

        private void buttonAllLightDim_Click(object sender, EventArgs e)
        {
            SendAllLightDim(comboBoxHouseCode.SelectedItem.ToString().ToCharArray()[0]);
        }

        private void buttonAllUnitOff_Click(object sender, EventArgs e)
        {
            SendAllUnitOff(comboBoxHouseCode.SelectedItem.ToString().ToCharArray()[0]);
        }


        public void SendAllLightOn(char houseCode)
        {
            SendCommand(houseCode, 0, Data.Commands.ALL_LIGHTS_ON, 1);
        }

        public void SendAllLightBright(char houseCode)
        {
            SendCommand(houseCode, 0, Data.Commands.BRIGHT, 1);
        }

        public void SendAllLightDim(char houseCode)
        {
            SendCommand(houseCode, 0, Data.Commands.DIM, 1);
        }

        public void SendAllUnitOff(char houseCode)
        {
            SendCommand(houseCode, 0, Data.Commands.ALL_UNITS_OFF, 1);
        }

        #endregion

        #region Server In/Out
        void CreateServerWriteEvent(char hcode, int ucode, byte fcode) //repeatCount
        {
            string subType = Data.Types.Power;
            int data = 0;

            switch (fcode)
            {
                case Data.Commands.DIM:
                    subType = Data.Types.Bright;
                    data = -1;
                    break;
                case Data.Commands.BRIGHT:
                    subType = Data.Types.Bright;
                    data = 1;
                    break;
                case Data.Commands.ALL_UNITS_OFF:
                case Data.Commands.OFF:
                    subType = Data.Types.Power;
                    data = 0;
                    break;
                case Data.Commands.ALL_LIGHTS_ON:
                case Data.Commands.ON:
                    subType = Data.Types.Power;
                    data = 1;
                    break;
            }

            Data.Base b = new Data.Base(Data.States.In, hcode, ucode, subType, data);
            string[] args = b.ToArray();

            MobiServer.ServerEventArgs e = new MobiServer.ServerEventArgs(args);//Data
            serverWriteHandler.Invoke(this, e);

        }

        override public void OnServerReadEvent(object sender, MobiServer.ServerEventArgs e)
        {
            try
            {
                Data.Base b = Data.Base.Parse(e.Args);

                if (b != null)
                {
                    if (b.State == Data.States.Attach)
                    {
                        StartControl(new Base(b.HouseCode, b.UnitCode, b.Type));
                    }
                }
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
        }
        
        #endregion

        #region Add Controls

        public void StartControl(Base control)
        {
            control.messageHandler += new MobiServer.MessageHandler(messageControl.OnMessageEvent);
            control.serverWriteHandler += new MobiServer.WriteHandler(serverControl.OnServerWriteEvent);
            serverControl.AddReadHandler(new MobiServer.ReadHandler(control.OnServerReadEvent));
            //serverControl.messageHandler += new MobiServer.MessageHandler(messageControl.OnMessageEvent);
            control.serialReadHandler += new MobiServer.Devices.X10.Data.ReadHandler(control.OnSerialReadEvent);
            control.serialWriteHandler += new MobiServer.Devices.X10.Data.WriteHandler(OnSerialWriteEvent);
            control.Start();
            AddControl(control);
            control.Dettach(false);
            
        }

        internal void WriteServerSockets(string[] args)
        {
            serverControl.WriteAllSockets(args);
        }

        public void AddControl(Base control)
        {
            Hashtable controlLine = null;
           bool addCompleted = false;
            
            for (int i = 0; i < controlHolder.Count; i++)
            {
                controlLine = (Hashtable)controlHolder[i];
                for (int j = 0; j < controlLine.Count; j++)
                {
                    Base c = (Base)controlLine[j];

                    if(control.CheckAddress(c.HouseCode,c.UnitCode))
                    {
                        int num = this.Controls.IndexOf(c);
                        if (c.DeviceType != control.DeviceType)
                        {
                            c.SetDeviceType(control.DeviceType);
                        }
                        this.Controls[num].Enabled = true;
                        this.Refresh();
                        return;
                    }
                }
            }

            for (int i = 0; i < controlHolder.Count; i++)
            {
                int lineWidth = 0;
                int y = 0;
                int lineHeight = 0;

                controlLine = (Hashtable)controlHolder[i];
                for (int j = 0; j < controlLine.Count; j++)
                {
                    Base c = (Base)controlLine[j];

                    if (!addCompleted)
                    {
                        lineWidth += c.Width;
                        y = c.Location.Y;
                        if (lineHeight < c.Height)
                            lineHeight = c.Height;
                    }
                    else
                    {
                        c.Location = new Point(c.Location.X, c.Location.Y + targetHeight);
                        controlLine[j] = c;
                    }
                }
                if (addCompleted)
                {
                    controlHolder[i] = controlLine;
                }
                else
                {
                    if (lineWidth + control.Width < this.Width)
                    {
                        control.Location = new Point(lineWidth, y);
                        targetControl = control;

                        if (lineHeight < control.Height)
                            targetHeight = control.Height - lineHeight;
                        else
                            targetHeight = 0;

                        this.Invoke(new EventHandler(AddControl));
                        controlLine.Add(controlLine.Count, control);
                        controlHolder[i] = controlLine;
                        addCompleted = true;
                    }
                }
            }

            if (!addCompleted)
            {
                control.Location = new Point(0, this.Height);// - 30
                targetControl = control;
                targetHeight = targetControl.Height;
                this.Invoke(new EventHandler(AddControl));               
                controlLine = new Hashtable();
                controlLine.Add(controlLine.Count, control);
                controlHolder.Add(controlHolder.Count, controlLine);
            }
            return;
        }

        void AddControl(object sender, EventArgs e)
        {
            this.SuspendLayout();
            this.Controls.Add(targetControl);
            if (CheckScreenHeight(this.Height + targetHeight + this.Location.Y))
                this.Height += targetHeight;
            else
                this.Height = System.Windows.Forms.Screen.PrimaryScreen.WorkingArea.Height - this.Location.Y;

            this.ResumeLayout(true);
        }

        bool CheckScreenHeight(int height)
        {
            if (height <= System.Windows.Forms.Screen.PrimaryScreen.WorkingArea.Height)
            {
                return true;
            }
            else
            {
                return false;
            }
        }

        private void buttonAttach_Click(object sender, EventArgs e)
        {
            try
            {
                char houseCode = comboBoxHouseCode.SelectedItem.ToString().ToCharArray()[0];
                int unitCode = int.Parse(comboBoxUnitCode.SelectedItem.ToString());
                string deviceType = comboBoxType.SelectedItem.ToString();
                switch (deviceType)
                {
                    case Data.Types.Appliance:
                        //AddControl(new Appliance(houseCode, unitCode));
                        StartControl(new Base(houseCode, unitCode, Data.Types.Appliance));

                        break;
                    case Data.Types.Lamp:
                        StartControl(new Base(houseCode, unitCode, Data.Types.Lamp));
                        break;
                    default:
                        throw new Exception(STR_ERROR_DEVICETYPE);
                }

                Data.Base b = new Data.Base(Data.States.Attach, houseCode, unitCode, deviceType);
                string[] args = b.ToArray();
                serverWriteHandler.Invoke(this, new MobiServer.ServerEventArgs(args));
            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
        }

        #endregion

        #region Save & Load

        public void SaveControls()
        {
            foreach (Hashtable controlLine in controlHolder.Values)
            {
                foreach (Base control in controlLine.Values)
                {
                    control.SaveSetting();

                }
            }

        }

        public override void SaveSetting()
        {
            StreamWriter fileStream = null;
            string file = System.Windows.Forms.Application.StartupPath + @"\" + SETTING_FILE;
            try
            {
                fileStream = new StreamWriter(file, false);

                fileStream.WriteLine("{0}\t{1}\t{2}", comboBoxComPort.SelectedItem.ToString(), checkBoxAutoStart.Checked, checkBoxAutoPortScan.Checked);
                fileStream.WriteLine("{0}\t{1}\t{2}", comboBoxHouseCode.SelectedItem.ToString(), comboBoxUnitCode.SelectedItem.ToString(), comboBoxType.SelectedItem.ToString());
            }
            catch (Exception ex)
            {
                PrintMessage(ex.Message);
            }
            finally
            {
                if (fileStream != null)
                {
                    fileStream.Close();
                    fileStream = null;
                }
            }
        }

        public override void LoadSetting()
        {
            StreamReader fileStream = null;
            string file = System.Windows.Forms.Application.StartupPath + @"\" + SETTING_FILE;
            string line = "";
            string[] args = null;

            try
            {
                if (!File.Exists(file))
                    return;

                fileStream = new StreamReader(file);

                line = fileStream.ReadLine();
                args = line.Split("\t".ToCharArray());
                comboBoxComPort.Text = args[0];
                checkBoxAutoStart.Checked = bool.Parse(args[1]);
                checkBoxAutoPortScan.Checked = bool.Parse(args[2]);

                line = fileStream.ReadLine();
                args = line.Split("\t".ToCharArray());
                comboBoxHouseCode.Text = args[0];
                comboBoxUnitCode.Text = (args[1]);
                comboBoxType.Text = (args[2]);

            }
            catch (Exception ex)
            {
                PrintError(ex.Message);
            }
            finally
            {
                if (fileStream != null)
                {
                    fileStream.Close();
                    fileStream = null;
                }
            }
        }

        #endregion
    }
}

